//*******************************************************************
//* A Custon BLE Design Using the Raytac MDBT50Q Module
//* REV 1.0.0
//* LAST UPDATE 04-05-2024
//* NOTES:
//*******************************************************************
#ifndef CECSVC_H_
#define CECSVC_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <zephyr/types.h>

// Define the 128 bit UUIDs for the GATT service and its characteristics 
// @brief SENSOR Service UUID. 
#define BT_UUID_SENSORSVC_VAL	\
	BT_UUID_128_ENCODE(0x97252dff,0x72ca,0x44ea,0x910b,0xa82864cae65f3)

// @brief SENSOR Characteristic. 
#define BT_UUID_SENSORCHR_VAL	\
	BT_UUID_128_ENCODE(0xa64a76ef,0xd522,0x4783,0x9f46,0xbfdf95f4510d)

#define BT_UUID_SENSORSVC	BT_UUID_DECLARE_128(BT_UUID_SENSORSVC_VAL)
#define BT_UUID_SENSORCHR	BT_UUID_DECLARE_128(BT_UUID_SENSORCHR_VAL)

// @brief Callback type for when SENSOR value change is received.                                          
typedef uint16_t (*sensorchrc_cb_t)(void);

// @brief Callback struct used by the SENSOR Service.                                         
struct sensorsvc_cb {
	// sensorchrc callback. 
	sensorchrc_cb_t sensorchrc_cb;
};

// Define the 128 bit UUIDs for the GATT service and its characteristics in */
/** @brief LED Service UUID. */
#define BT_UUID_LEDSVC_VAL		\
	BT_UUID_128_ENCODE(0x5328b08a,0x0c4a,0x4bda,0x996c,0x41064568e694)

/** @brief LED Control Characteristic. */
#define BT_UUID_LEDCHR_VAL		\
	BT_UUID_128_ENCODE(0x9488b3b2,0x7de5,0x4a1c,0xaf41,0x385a89931332)

#define BT_UUID_LEDSVC		BT_UUID_DECLARE_128(BT_UUID_LEDSVC_VAL)
#define BT_UUID_LEDCHR		BT_UUID_DECLARE_128(BT_UUID_LEDCHR_VAL)

/** @brief Callback type for when LED status change is received. */
typedef void (*ledchrc_cb_t)(uint8_t ledcntlbite);

/** @brief Callback struct used by the LED Service. */
struct ledsvc_cb {
	/** LEDCHR callback. */
	ledchrc_cb_t ledchrc_cb;
};

/** @brief Initialize the SENSOR Service.
 *
 * This function registers application callback functions with the SENSOR
 * Service
 *
 * @param[in] sensor_callback Struct containing pointers to callback functions
 *			used by the service. This pointer can be NULL
 *			if no callback functions are defined.
 *
 *
 * @retval 0 If the operation was successful.
 *           Otherwise, a (negative) error code is returned.
 */
int sensorsvc_init(struct sensorsvc_cb *sensor_callback);

/** @brief Initialize the LED Service.
 *
 * This function registers application callback functions with the LED
 * Service
 *
 * @param[in] led_callback Struct containing pointers to callback functions
 *			used by the service. This pointer can be NULL
 *			if no callback functions are defined.
 *
 *
 * @retval 0 If the operation was successful.
 *           Otherwise, a (negative) error code is returned.
 */
int ledsvc_init(struct ledsvc_cb *led_callback);

#ifdef __cplusplus
}
#endif

#endif
