#!/usr/bin/python3
#
# Based on code found within the Bluetooth Linux Study Guide
# Modified by Fred Eady

from gi.repository import GLib

import bluetooth_utils
import bluetooth_constants
import dbus
import dbus.mainloop.glib
import sys
import time
sys.path.insert(0, '.')
  
adapter_interface = None
mainloop = None
timer_id = None

devices = {}
managed_objects_found = 0
devname = None
bdaddr = None
bdaddr_chosen = None
scansecs = 20
device_interface = None
bus = None

device_path = None
found_sensorsvc = False
found_sensorchrc  = False
sensorsvc_path = None
sensorchrc_path  = None

found_ledsvc = False
found_ledchrc  = False
ledsvc_path = None
ledchrc_path  = None

def sensor_data_received(interface, changed, invalidated, path):
    if 'Value' in changed:
        new_sensor_val = bluetooth_utils.dbus_to_python(changed['Value'])
        print(new_sensor_val)
        
def start_notifications():
    global sensorchrc_path
    global bus
    char_proxy = bus.get_object(bluetooth_constants.BLUEZ_SERVICE_NAME,sensorchrc_path)
    char_interface = dbus.Interface(char_proxy,bluetooth_constants.GATT_CHARACTERISTIC_INTERFACE)
    
    bus.add_signal_receiver(sensor_data_received,
        dbus_interface = bluetooth_constants.DBUS_PROPERTIES,
        signal_name = "PropertiesChanged",
        path = sensorchrc_path,
        path_keyword = "path")
    
    try:
        print("Starting notifications")
        char_interface.StartNotify()
        print("Done starting notifications")
    except Exception as e:
        print("Fialed to start sensor data notificationa")
        print(e.get_dbus_name())
        print(e.get_dbus_message())
        return bluetooth_constants.RESULT_EXCEPTION
    else:
        return bluetooth_constants.RESULT_OK
    
def led_on():
    global ledchrc_path
    char_proxy = bus.get_object(bluetooth_constants.BLUEZ_SERVICE_NAME,ledchrc_path)
    char_interface = dbus.Interface(char_proxy,bluetooth_constants.GATT_CHARACTERISTIC_INTERFACE)
    try:
        ledcmd = int(1).to_bytes(1, byteorder='little')
        char_interface.WriteValue(ledcmd,{})
    except Exception as e:
        print("Failed to write to LED")
        print(e.get_dbus_name())
        print(e.get_dbus_message())
        return bluetooth_constants.RESULT_EXCEPTION
    else:
        print("LED ON command sent")
        return bluetooth_constants.RESULT_OK

def led_off():
    global ledchrc_path
    char_proxy = bus.get_object(bluetooth_constants.BLUEZ_SERVICE_NAME,ledchrc_path)
    char_interface = dbus.Interface(char_proxy,bluetooth_constants.GATT_CHARACTERISTIC_INTERFACE)
    try:
        ledcmd = int(0).to_bytes(1, byteorder='little')
        char_interface.WriteValue(ledcmd,{})
    except Exception as e:
        print("Failed to write to LED")
        print(e.get_dbus_name())
        print(e.get_dbus_message())
        return bluetooth_constants.RESULT_EXCEPTION
    else:
        print("LED OFF command sent")
        return bluetooth_constants.RESULT_OK
        
    
def read_sensor():
    global sensorchrc_path
    char_proxy = bus.get_object(bluetooth_constants.BLUEZ_SERVICE_NAME,sensorchrc_path)
    char_interface = dbus.Interface(char_proxy,bluetooth_constants.GATT_CHARACTERISTIC_INTERFACE)
    
    try:
        value = char_interface.ReadValue({})
    except Exception as e:
        print("Failed to read sensor")
        print(e.get_dbus_name())
        print(e.get_dbus_message())
        return bluetooth_constants.RESULT_EXCEPTION
    else:
        sensorvalHI = bluetooth_utils.dbus_to_python(value[1])
        sensorvalLO = bluetooth_utils.dbus_to_python(value[0])
        sensorval = sensorvalHI << 8
        sensorval = sensorval + sensorvalLO
        print("Sensor Value in mv:",int(sensorval))
        return bluetooth_constants.RESULT_OK
        

def service_discovery_completed():
    global found_sensorsvc
    global found_sensorchrc
    global sensorsvc_path
    global sensorchrc_path
    global found_ledsvc
    global found_ledchrc
    global ledsvc_path
    global ledchrc_path
    global bus
    
    if found_sensorsvc and found_sensorchrc and found_ledsvc and found_ledchrc:
        print("Required services and characteristics found - device is OK")
        print("Sensor Service path: ",sensorsvc_path)
        print("Sensor Characteristic path: ",sensorchrc_path)
        print("LED Service path: ",ledsvc_path)
        print("LED Characteristic path: ",ledchrc_path)
        #read_sensor()
        led_off()
        time.sleep(2)
        led_on()
        time.sleep(2)
        led_off()
        #start_notifications()
    else:
        print("Required services and characteristics were not found - device is NOK")
        print("Sensor Service found: ",str(found_sensorsvc))
        print("Sensor Characteristic found: ",str(found_sensorchrc))
        print("LED Service found: ",str(found_ledsvc))
        print("LED Characteristic found: ",str(found_ledchrc))
        bus.remove_signal_receiver(interfaces_added,"InterfacesAdded")
        bus.remove_signal_receiver(properties_changed,"PropertiesChanged")
        mainloop.quit()

    bus.remove_signal_receiver(interfaces_added,"InterfacesAdded")
    bus.remove_signal_receiver(properties_changed,"PropertiesChanged")
 
def get_nrf52840_addr():
    global devname
    global bdaddr
    for path in devices:
        #print((bluetooth_utils.dbus_to_python(path)))
        dev = devices[path]
        if 'Name' in dev:
            devname = (bluetooth_utils.dbus_to_python(dev['Name']))
            #print(devname)
        if devname == 'NRF52840_BRD':
            if 'Address' in dev:
                bdaddr = (bluetooth_utils.dbus_to_python(dev['Address']))
            print("nRF52840 Device Name and Address:")
            print("------------------------------")
            print(devname)
            print(bdaddr)
            return 
               
def interfaces_added(path, interfaces):
    global found_sensorsvc
    global found_sensorchrc
    global sensorsvc_path
    global sensorchrc_path
    global found_ledsvc
    global found_ledchrc
    global ledsvc_path
    global ledchrc_path
    
    if bluetooth_constants.DEVICE_INTERFACE in interfaces:
        device_properties = interfaces[bluetooth_constants.DEVICE_INTERFACE]
        if path not in devices:
            print("NEW path  :", path)
            devices[path] = device_properties
            dev = devices[path]
            if 'Address' in dev:
                print("NEW bdaddr: ", bluetooth_utils.dbus_to_python(dev['Address']))
            if 'Name' in dev:
                print("NEW name  : ", bluetooth_utils.dbus_to_python(dev['Name']))
            if 'RSSI' in dev:
                print("NEW RSSI  : ", bluetooth_utils.dbus_to_python(dev['RSSI']))
                print("------------------------------")

    if bluetooth_constants.GATT_SERVICE_INTERFACE in interfaces:
        properties = interfaces[bluetooth_constants.GATT_SERVICE_INTERFACE]
        print("--------------------------------------------------------------------------------")
        print("SVC path   :", path)
        if 'UUID' in properties:
            uuid = properties['UUID']
            if uuid == bluetooth_constants.SENSOR_SVC_UUID:
                found_sensorsvc = True
                sensorsvc_path = path
                
            if uuid == bluetooth_constants.LED_SVC_UUID:
                found_ledsvc = True
                ledsvc_path = path
                
            print("SVC UUID   : ", bluetooth_utils.dbus_to_python(uuid))
            print("SVC name   : ", bluetooth_utils.get_name_from_uuid(uuid))
        
    if bluetooth_constants.GATT_CHARACTERISTIC_INTERFACE in interfaces:
        properties = interfaces[bluetooth_constants.GATT_CHARACTERISTIC_INTERFACE]
        print("  CHR path   :", path)
        if 'UUID' in properties:
            uuid = properties['UUID']
            if uuid == bluetooth_constants.SENSOR_CHR_UUID:
                found_sensorchrc = True
                sensorchrc_path = path
                
            if uuid == bluetooth_constants.LED_CHR_UUID:
                found_ledchrc = True
                ledchrc_path = path
                
            print("  CHR UUID   : ", bluetooth_utils.dbus_to_python(uuid))
            print("  CHR name   : ", bluetooth_utils.get_name_from_uuid(uuid))
            flags  = ""
            for flag in properties['Flags']:
                flags = flags + flag + ","
            print("  CHR flags  : ", flags)
    
    if bluetooth_constants.GATT_DESCRIPTOR_INTERFACE in interfaces:
        properties = interfaces[bluetooth_constants.GATT_DESCRIPTOR_INTERFACE]
        print("    DSC path   :", path)
        if 'UUID' in properties:
            uuid = properties['UUID']
            print("    DSC UUID   : ", bluetooth_utils.dbus_to_python(uuid))
            print("    DSC name   : ", bluetooth_utils.get_name_from_uuid(uuid))
        
def interfaces_removed(path, interfaces):
    # interfaces is an array of dictionary strings in this signal
    if not bluetooth_constants.DEVICE_INTERFACE in interfaces:
        return
    if path in devices:
        dev = devices[path]
        if 'Address' in dev:
            print("DEL bdaddr: ", bluetooth_utils.dbus_to_python(dev['Address']))
        else:
            print("DEL path  : ", path)
            print("------------------------------")
        del devices[path]

def properties_changed(interface, changed, invalidated, path):
    
    if 'ServicesResolved' in changed:
        sr = bluetooth_utils.dbus_to_python(changed['ServicesResolved'])
        print("ServicesResolved  : ", sr)
        if sr == True:
            service_discovery_completed()
        else:
            mainloop.quit()
    
    if 'RSSI' in changed:
        signalstrength = bluetooth_utils.dbus_to_python(changed['RSSI'])
        print("signalstrength",signalstrength)
        #if signalstrength == 0:
        #    mainloop.quit()

    if interface != bluetooth_constants.DEVICE_INTERFACE:
        #print("no bluetooth_constants.DEVICE_INTERFACE")
        return
    if path in devices:
        devices[path] = dict(devices[path].items())
        devices[path].update(changed.items())
    else:
        devices[path] = changed

    dev = devices[path]
    print("CHG path  :", path)
    if 'Address' in dev:
        print("CHG bdaddr: ", bluetooth_utils.dbus_to_python(dev['Address']))
    if 'Name' in dev:
        print("CHG name  : ", bluetooth_utils.dbus_to_python(dev['Name']))
    if 'RSSI' in dev:
        print("CHG RSSI  : ", bluetooth_utils.dbus_to_python(dev['RSSI']))
    print("------------------------------")

def get_known_devices(bus):
    global managed_objects_found
    object_manager = dbus.Interface(bus.get_object(bluetooth_constants.BLUEZ_SERVICE_NAME, "/"), bluetooth_constants.DBUS_OM_IFACE)
    managed_objects=object_manager.GetManagedObjects()

    for path, intrfaces in managed_objects.items():
        for intrface_name in intrfaces:
            if intrface_name == bluetooth_constants.DEVICE_INTERFACE:
                managed_objects_found += 1
                print("EXI path  : ", path)
                device_properties = intrfaces[bluetooth_constants.DEVICE_INTERFACE]
                devices[path] = device_properties
                if 'Address' in device_properties:
                    print("EXI bdaddr: ", bluetooth_utils.dbus_to_python(device_properties['Address']))
                if 'Connected' in device_properties:
                    print("EXI cncted: ", bluetooth_utils.dbus_to_python(device_properties['Connected']))           
                print("------------------------------")
            
def discovery_timeout():
    global adapter_interface
    global mainloop
    global timer_id
    GLib.source_remove(timer_id)
    mainloop.quit()
    adapter_interface.StopDiscovery()
    bus = dbus.SystemBus()
    bus.remove_signal_receiver(interfaces_added,"InterfacesAdded")
    bus.remove_signal_receiver(interfaces_added,"InterfacesRemoved")
    bus.remove_signal_receiver(properties_changed,"PropertiesChanged")
    get_nrf52840_addr()
    return True

def discover_devices(bus,timeout):
    global adapter_interface
    global mainloop
    global timer_id
    
    adapter_path = bluetooth_constants.BLUEZ_NAMESPACE + bluetooth_constants.ADAPTER_NAME

    # acquire the adapter interface so we can call its methods 
    adapter_object = bus.get_object(bluetooth_constants.BLUEZ_SERVICE_NAME, adapter_path)
    adapter_interface=dbus.Interface(adapter_object, bluetooth_constants.ADAPTER_INTERFACE)
    
    # register signal handler functions so we can asynchronously report discovered devices
    
    # InterfacesAdded signal is emitted by BlueZ when an advertising packet from a device it doesn't
    # already know about is received
    bus.add_signal_receiver(interfaces_added,
            dbus_interface = bluetooth_constants.DBUS_OM_IFACE,
            signal_name = "InterfacesAdded")

    # InterfacesRemoved signal is emitted by BlueZ when a device "goes away"
    bus.add_signal_receiver(interfaces_removed,
            dbus_interface = bluetooth_constants.DBUS_OM_IFACE,
            signal_name = "InterfacesRemoved")

    # PropertiesChanged signal is emitted by BlueZ when something re: a device already encountered
    # changes e.g. the RSSI value
    bus.add_signal_receiver(properties_changed,
           dbus_interface = bluetooth_constants.DBUS_PROPERTIES,
            signal_name = "PropertiesChanged",
            path_keyword = "path")
    
    mainloop = GLib.MainLoop()
    timer_id = GLib.timeout_add(timeout, discovery_timeout)
    adapter_interface.StartDiscovery(byte_arrays=True)

    mainloop.run()
    
    device_list = devices.values()
    discovered_devices = []
    for device in device_list:
        dev = {}
        discovered_devices.append(dev)

    return discovered_devices

def connect():
    global bus
    global device_interface
    try:
        device_interface.Connect()
    except Exception as e:
        print("Failed to connect")
        print(e.get_dbus_name())
        print(e.get_dbus_message())
        if ("UnknownObject" in e.get_dbus_name()):
            print("Try scanning first to resolve this problem")
        return bluetooth_constants.RESULT_EXCEPTION
    else:
        print("Connected OK")
        return bluetooth_constants.RESULT_OK

def disconnect():
    global bus
    global device_interface
    try:
        device_interface.Disconnect()
    except Exception as e:
        print("Failed to disconnect")
        print(e.get_dbus_name())
        print(e.get_dbus_message())
        return bluetooth_constants.RESULT_EXCEPTION
    else:
        print("Disconnected OK")
        return bluetooth_constants.RESULT_OK

if (len(sys.argv) != 2):
    # our device name = NRF52840_BRD
    print("usage: python3 pi5_ble_central_client.py [device name]")
    sys.exit(1)
    
# this is the time in seconds that device discovery will be active
scantime = scansecs * 1000

# dbus initialisation steps
dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
bus = dbus.SystemBus()

# ask for a list of devices already known to the BlueZ daemon
print("Listing devices already known to BlueZ:")
get_known_devices(bus)
print("Found ",managed_objects_found," managed device objects")
print("Scanning")
discover_devices(bus, scantime)
    
bdaddr_chosen = sys.argv[1]
bus = dbus.SystemBus()
adapter_path = bluetooth_constants.BLUEZ_NAMESPACE + bluetooth_constants.ADAPTER_NAME
device_path = bluetooth_utils.device_address_to_path(bdaddr, adapter_path)
device_proxy = bus.get_object(bluetooth_constants.BLUEZ_SERVICE_NAME,device_path)
device_interface = dbus.Interface(device_proxy, bluetooth_constants.DEVICE_INTERFACE)

print("Connecting to " + bdaddr)
connect()
print("The Bluetooth stack is automatically discovering services")
print("Registering to receive InterfacesAdded signals")
bus.add_signal_receiver(interfaces_added,
        dbus_interface = bluetooth_constants.DBUS_OM_IFACE,
        signal_name = "InterfacesAdded")
print("Registering to receive PropertiesChanged signals")
bus.add_signal_receiver(properties_changed,
        dbus_interface = bluetooth_constants.DBUS_PROPERTIES,
        signal_name = "PropertiesChanged",
        path_keyword = "path")
mainloop = GLib.MainLoop()
mainloop.run()
print("Finished")
print("Disconnecting from " + bdaddr)
disconnect()

